package checker;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;

/**
 *
 * @author Rick Hendricksen
 */
public class FileChecker {
    final static String[] keys = new String[]{
        "GL_MAP1_VERTEX_",
        "GL_MAP1_INDEX",
        "GL_MAP1_COLOR_4",
        "GL_MAP1_NORMAL",
        "GL_MAP1_TEXTURE_COORD_",
        "glEvalCoord1f",
        "glEvalCoord2f",
        "glEvalCoord1d",
        "glEvalCoord2f",
        "glMap1f",
        "glMap2f",
        "glMap1d",
        "glMap2d",
        "GL_MAP2_VERTEX_",
        "GL_MAP2_INDEX",
        "GL_MAP2_COLOR_4",
        "GL_MAP2_NORMAL",
        "GL_MAP2_TEXTURE_COORD_",
        "GL_AUTO_NORMAL",
        "GLUquadric",
        "GL_TEXTURE_GEN_S", 
        "GL_TEXTURE_GEN_T",
        "glCallList",
        "glGenLists",
        "glNewList",
        "glEndList",
        "glDeleteLists",
        "GL_POLYGON_SMOOTH",
        "GL_LINE_SMOOTH",
        "GL_POINT_SMOOTH",
        "glLightModelfv", // use gl.glLightfv(GL_LIGHT0, GL_AMBIENT, new float[]{0.1f, 0.1f, 0.1f, 1f}, 0); (or something similar)
    };
    
    
    /**
     * This function checks if an illegal substring is used in a java file. 
     * It ignores keys in strings and comments.
     * 
     * @param br
     * @param filename
     * @return true if an illegal function was found
     * @throws IOException 
     */
    static boolean check(BufferedReader br, String filename) throws IOException{
        boolean ret = false;
        
        boolean multilineComment = false;
        boolean singleLineComment = false;
        boolean star = false;
        boolean slash = false;
        
        int line = 1;
        int c;
        
        StringBuilder sb = new StringBuilder();
        
        while((c = br.read()) != -1){
            
            if(c == '\n') {
                String code = sb.toString();
                
                for(String s : keys){
                    if(code.contains(s)){
                        String warning = "Illegal use of \""+s+"\" in " + filename + " on line " + line + ".";
                        System.err.println(warning);
                        System.exit(0);
                        ret = true;
                    }
                }
                
                sb.delete(0, sb.length());
                line++;
                singleLineComment = false;
            }
            
            if(multilineComment){
                if(star && c == '/'){
                    multilineComment = false;
                    star = false;
                }else {
                    star = c == '*';
                }
            }
            
            if(multilineComment || singleLineComment){
                continue;
            }
            
            switch(c){
                case '/':
                    if(slash){
                        singleLineComment = true;
                        slash = false;
                    }else slash = true;
                    break;
                case '*':
                    if(slash){
                        multilineComment = true;
                    }
                    slash = false;
                    break;
                case '"':
                    slash = false;
                    skipString(br);
                    break;
                default:
                    slash = false;
                    sb.append((char) c);
            }
        }
        return ret;
    }

    private static void skipString(BufferedReader br) throws IOException {
        int c;
        while((c = br.read()) != -1){
            if(c == '\\'){
                br.read(); // skip next character
                continue;
            }
            if(c == '"'){
                return;
            }
        }
    }
    
    public static void testDir(File dir, String prefix){
        File[] listFiles = dir.listFiles();
        for(File f : listFiles){
            if(f.isDirectory()){
                testDir(f, prefix + f.getName() + "/");
            }else{
                if(f.getName().endsWith(".java")){
                    try {
                        test(new FileInputStream(f), prefix + f.getName());
                    } catch (IOException ex) {
                        ex.printStackTrace();
                    }
                }
            }
        }
    }
    
    public static void test(InputStream in, String filename) throws IOException{
        try(BufferedReader br = new BufferedReader(new InputStreamReader(in))){
            check(br, filename);
        }
    }
}
